@@ -8,21 +8,30 @@ require "sentry/version"
 require "sentry/exceptions"
 require "sentry/core_ext/object/deep_dup"
 require "sentry/utils/argument_checking_helper"
+require "sentry/utils/encoding_helper"
 require "sentry/utils/logging_helper"
+require "sentry/utils/sample_rand"
 require "sentry/configuration"
-require "sentry/logger"
+require "sentry/structured_logger"
+require "sentry/debug_structured_logger"
 require "sentry/event"
 require "sentry/error_event"
 require "sentry/transaction_event"
+require "sentry/check_in_event"
 require "sentry/span"
 require "sentry/transaction"
 require "sentry/hub"
 require "sentry/background_worker"
+require "sentry/threaded_periodic_worker"
 require "sentry/session_flusher"
+require "sentry/backpressure_monitor"
+require "sentry/cron/monitor_check_ins"
+require "sentry/vernier/profiler"
+require "sentry/metrics"
 
 [
   "sentry/rake",
-  "sentry/rack",
+  "sentry/rack"
 ].each do |lib|
   begin
     require lib
@@ -35,12 +44,24 @@ module Sentry
 
   CAPTURED_SIGNATURE = :@__sentry_captured
 
-  LOGGER_PROGNAME = "sentry".freeze
+  LOGGER_PROGNAME = "sentry"
 
-  SENTRY_TRACE_HEADER_NAME = "sentry-trace".freeze
+  SENTRY_TRACE_HEADER_NAME = "sentry-trace"
+
+  BAGGAGE_HEADER_NAME = "baggage"
 
   THREAD_LOCAL = :sentry_hub
 
+  MUTEX = Mutex.new
+
+  GLOBALS = %i[
+    main_hub
+    logger
+    session_flusher
+    backpressure_monitor
+    exception_locals_tp
+  ].freeze
+
   class << self
     # @!visibility private
     def exception_locals_tp
@@ -64,26 +85,40 @@ module Sentry
     attr_accessor :background_worker
 
     # @!attribute [r] session_flusher
-    #   @return [SessionFlusher]
+    #   @return [SessionFlusher, nil]
     attr_reader :session_flusher
 
+    # @!attribute [r] backpressure_monitor
+    #   @return [BackpressureMonitor, nil]
+    attr_reader :backpressure_monitor
+
     ##### Patch Registration #####
 
     # @!visibility private
-    def register_patch(&block)
-      registered_patches << block
+    def register_patch(key, patch = nil, target = nil, &block)
+      if patch && block
+        raise ArgumentError.new("Please provide either a patch and its target OR a block, but not both")
+      end
+
+      if block
+        registered_patches[key] = block
+      else
+        registered_patches[key] = proc do
+          target.send(:prepend, patch) unless target.ancestors.include?(patch)
+        end
+      end
     end
 
     # @!visibility private
     def apply_patches(config)
-      registered_patches.each do |patch|
-        patch.call(config)
+      registered_patches.each do |key, patch|
+        patch.call(config) if config.enabled_patches.include?(key)
       end
     end
 
     # @!visibility private
     def registered_patches
-      @registered_patches ||= []
+      @registered_patches ||= {}
     end
 
     ##### Integrations #####
@@ -101,7 +136,7 @@ module Sentry
     # @param version [String] version of the integration
     def register_integration(name, version)
       if initialized?
-        logger.warn(LOGGER_PROGNAME) do
+        sdk_logger.warn(LOGGER_PROGNAME) do
           <<~MSG
             Integration '#{name}' is loaded after the SDK is initialized, which can cause unexpected behavior.  Please make sure all integrations are loaded before SDK initialization.
           MSG
@@ -185,6 +220,13 @@ module Sentry
       get_current_scope.set_context(*args)
     end
 
+    # @!method add_attachment
+    #   @!macro add_attachment
+    def add_attachment(**opts)
+      return unless initialized?
+      get_current_scope.add_attachment(**opts)
+    end
+
     ##### Main APIs #####
 
     # Initializes the SDK with given configuration.
@@ -192,30 +234,52 @@ module Sentry
     # @yieldparam config [Configuration]
     # @return [void]
     def init(&block)
-      config = Configuration.new
-      yield(config) if block_given?
+      config = Configuration.new(&block)
+
       config.detect_release
       apply_patches(config)
+      config.validate
       client = Client.new(config)
       scope = Scope.new(max_breadcrumbs: config.max_breadcrumbs)
       hub = Hub.new(client, scope)
       Thread.current.thread_variable_set(THREAD_LOCAL, hub)
       @main_hub = hub
       @background_worker = Sentry::BackgroundWorker.new(config)
+      @session_flusher = config.session_tracking? ? Sentry::SessionFlusher.new(config, client) : nil
+      @backpressure_monitor = config.enable_backpressure_handling ? Sentry::BackpressureMonitor.new(config, client) : nil
+      exception_locals_tp.enable if config.include_local_variables
+      at_exit { close }
+    end
 
-      @session_flusher = if config.auto_session_tracking
-                           Sentry::SessionFlusher.new(config, client)
-                         else
-                           nil
-                         end
-
-      if config.capture_exception_frame_locals
-        exception_locals_tp.enable
+    # Flushes pending events and cleans up SDK state.
+    # SDK will stop sending events and all top-level APIs will be no-ops after this.
+    #
+    # @return [void]
+    def close
+      if @session_flusher
+        @session_flusher.flush
+        @session_flusher.kill
+        @session_flusher = nil
       end
 
-      at_exit do
-        @session_flusher&.kill
-        @background_worker.shutdown
+      if @backpressure_monitor
+        @backpressure_monitor.kill
+        @backpressure_monitor = nil
+      end
+
+      if client = get_current_client
+        client.flush
+
+        if client.configuration.include_local_variables
+          exception_locals_tp.disable
+        end
+      end
+
+      @background_worker.shutdown
+
+      MUTEX.synchronize do
+        @main_hub = nil
+        Thread.current.thread_variable_set(THREAD_LOCAL, nil)
       end
     end
 
@@ -243,6 +307,9 @@ module Sentry
     #
     # @return [Hub]
     def get_main_hub
+      MUTEX.synchronize { @main_hub }
+    rescue ThreadError
+      # In some rare cases this may be called in a trap context so we need to handle it gracefully
       @main_hub
     end
 
@@ -287,6 +354,7 @@ module Sentry
     #
     # @return [void]
     def clone_hub_to_current_thread
+      return unless initialized?
       Thread.current.thread_variable_set(THREAD_LOCAL, get_main_hub.clone)
     end
 
@@ -327,7 +395,7 @@ module Sentry
     # @yieldparam scope [Scope]
     # @return [void]
     def with_scope(&block)
-      return unless initialized?
+      return yield unless initialized?
       get_current_hub.with_scope(&block)
     end
 
@@ -360,6 +428,25 @@ module Sentry
       get_current_hub.capture_exception(exception, **options, &block)
     end
 
+    # Takes a block and evaluates it. If the block raised an exception, it reports the exception to Sentry and re-raises it.
+    # If the block ran without exception, it returns the evaluation result.
+    #
+    # @example
+    #   Sentry.with_exception_captured do
+    #     1/1 #=> 1 will be returned
+    #   end
+    #
+    #   Sentry.with_exception_captured do
+    #     1/0 #=> ZeroDivisionError will be reported and re-raised
+    #   end
+    #
+    def with_exception_captured(**options, &block)
+      yield
+    rescue Exception => e
+      capture_exception(e, **options)
+      raise
+    end
+
     # Takes a message string and reports it to Sentry via the currently active hub.
     #
     # @yieldparam scope [Scope]
@@ -377,6 +464,42 @@ module Sentry
       get_current_hub.capture_event(event)
     end
 
+    # Captures a check-in and sends it to Sentry via the currently active hub.
+    #
+    # @param slug [String] identifier of this monitor
+    # @param status [Symbol] status of this check-in, one of {CheckInEvent::VALID_STATUSES}
+    #
+    # @param [Hash] options extra check-in options
+    # @option options [String] check_in_id for updating the status of an existing monitor
+    # @option options [Integer] duration seconds elapsed since this monitor started
+    # @option options [Cron::MonitorConfig] monitor_config configuration for this monitor
+    #
+    # @return [String, nil] The {CheckInEvent#check_in_id} to use for later updates on the same slug
+    def capture_check_in(slug, status, **options)
+      return unless initialized?
+      get_current_hub.capture_check_in(slug, status, **options)
+    end
+
+    # Captures a log event and sends it to Sentry via the currently active hub.
+    # This is the underlying method used by the StructuredLogger class.
+    #
+    # @param message [String] the log message
+    # @param [Hash] options Extra log event options
+    # @option options [Symbol] level The log level (:trace, :debug, :info, :warn, :error, :fatal)
+    # @option options [Integer] severity The severity number according to the Sentry Logs Protocol
+    # @option options [String] origin The origin of the log event (e.g., "auto.db.rails", "manual")
+    # @option options [Hash] Additional attributes to include with the log
+    #
+    # @example Direct usage (prefer using Sentry.logger instead)
+    #   Sentry.capture_log("User logged in", level: :info, user_id: 123)
+    #
+    # @see https://develop.sentry.dev/sdk/telemetry/logs/ Sentry SDK Telemetry Logs Protocol
+    # @return [LogEvent, nil] The created log event or nil if logging is disabled
+    def capture_log(message, **options)
+      return unless initialized?
+      get_current_hub.capture_log_event(message, **options)
+    end
+
     # Takes or initializes a new Sentry::Transaction and makes a sampling decision for it.
     #
     # @return [Transaction, nil]
@@ -399,24 +522,8 @@ module Sentry
     #   end
     #
     def with_child_span(**attributes, &block)
-      current_span = get_current_scope.get_span
-
-      if current_span
-        result = nil
-
-        begin
-          current_span.with_child_span(**attributes) do |child_span|
-            get_current_scope.set_span(child_span)
-            result = yield(child_span)
-          end
-        ensure
-          get_current_scope.set_span(current_span)
-        end
-
-        result
-      else
-        yield(nil)
-      end
+      return yield(nil) unless Sentry.initialized?
+      get_current_hub.with_child_span(**attributes, &block)
     end
 
     # Returns the id of the lastly reported Sentry::Event.
@@ -435,6 +542,110 @@ module Sentry
       !!exc.instance_variable_get(CAPTURED_SIGNATURE)
     end
 
+    # Add a global event processor [Proc].
+    # These run before scope event processors.
+    #
+    # @yieldparam event [Event]
+    # @yieldparam hint [Hash, nil]
+    # @return [void]
+    #
+    # @example
+    #   Sentry.add_global_event_processor do |event, hint|
+    #     event.tags = { foo: 42 }
+    #     event
+    #   end
+    #
+    def add_global_event_processor(&block)
+      Scope.add_global_event_processor(&block)
+    end
+
+    # Returns the traceparent (sentry-trace) header for distributed tracing.
+    # Can be either from the currently active span or the propagation context.
+    #
+    # @return [String, nil]
+    def get_traceparent
+      return nil unless initialized?
+      get_current_hub.get_traceparent
+    end
+
+    # Returns the baggage header for distributed tracing.
+    # Can be either from the currently active span or the propagation context.
+    #
+    # @return [String, nil]
+    def get_baggage
+      return nil unless initialized?
+      get_current_hub.get_baggage
+    end
+
+    # Returns the a Hash containing sentry-trace and baggage.
+    # Can be either from the currently active span or the propagation context.
+    #
+    # @return [Hash, nil]
+    def get_trace_propagation_headers
+      return nil unless initialized?
+      get_current_hub.get_trace_propagation_headers
+    end
+
+    # Returns the a Hash containing sentry-trace and baggage.
+    # Can be either from the currently active span or the propagation context.
+    #
+    # @return [String]
+    def get_trace_propagation_meta
+      return "" unless initialized?
+      get_current_hub.get_trace_propagation_meta
+    end
+
+    # Continue an incoming trace from a rack env like hash.
+    #
+    # @param env [Hash]
+    # @return [Transaction, nil]
+    def continue_trace(env, **options)
+      return nil unless initialized?
+      get_current_hub.continue_trace(env, **options)
+    end
+
+    # Returns the structured logger instance that implements Sentry's SDK telemetry logs protocol.
+    #
+    # This logger is only available when logs are enabled in the configuration.
+    #
+    # @example Enable logs in configuration
+    #   Sentry.init do |config|
+    #     config.dsn = "YOUR_DSN"
+    #     config.enable_logs = true
+    #   end
+    #
+    # @example Basic usage
+    #   Sentry.logger.info("User logged in successfully", user_id: 123)
+    #   Sentry.logger.error("Failed to process payment",
+    #     transaction_id: "tx_123",
+    #     error_code: "PAYMENT_FAILED"
+    #   )
+    #
+    # @see https://develop.sentry.dev/sdk/telemetry/logs/ Sentry SDK Telemetry Logs Protocol
+    #
+    # @return [StructuredLogger] The structured logger instance or nil if logs are disabled
+    def logger
+      @logger ||= configuration.structured_logging.logger_class.new(configuration)
+    end
+
+    # Returns the metrics API for capturing custom metrics.
+    #
+    # @example Enable metrics
+    #   Sentry.init do |config|
+    #     config.dsn = "YOUR_DSN"
+    #     config.enable_metrics = true
+    #   end
+    #
+    # @example Usage
+    #   Sentry.metrics.count("button.click", 1, attributes: { button_id: "submit" })
+    #   Sentry.metrics.distribution("response.time", 120.5, unit: "millisecond")
+    #   Sentry.metrics.gauge("cpu.usage", 75.2, unit: "percent")
+    #
+    # @return [Metrics] The metrics API
+    def metrics
+      Metrics
+    end
+
     ##### Helpers #####
 
     # @!visibility private
@@ -446,8 +657,8 @@ module Sentry
     end
 
     # @!visibility private
-    def logger
-      configuration.logger
+    def sdk_logger
+      configuration.sdk_logger
     end
 
     # @!visibility private
@@ -459,9 +670,19 @@ module Sentry
     def utc_now
       Time.now.utc
     end
+
+    # @!visibility private
+    def dependency_installed?(name)
+      Object.const_defined?(name)
+    end
   end
 end
 
 # patches
 require "sentry/net/http"
 require "sentry/redis"
+require "sentry/puma"
+require "sentry/graphql"
+require "sentry/faraday"
+require "sentry/excon"
+require "sentry/std_lib_logger"
